local cameraBorder = 300
local screenWidth, screenHeight = love.graphics.getDimensions()
local drawCanvas = love.graphics.newCanvas(screenWidth - cameraBorder, screenHeight)
local displayCanvas = love.graphics.newCanvas(screenWidth - cameraBorder, screenHeight)
local tileWidth = 64
local tileHeight = tileWidth * .5
local halfWidth = tileHeight
local halfHeight = tileHeight * .5
local offsetX = 1720
local offsetY = 700
local scrollSpeed = 20
local scale = .25
local zoomSpeed = .01

local octaves = 4
local texture = engine.texture[engine.currentPlanetTexture]

local waterLevel = texture[1].height
local sandLevel = texture[2].height
local dirtLevel = texture[3].height
local grassLevel = texture[4].height
local mountainLevel = texture[5].height

local treeChance = 50

local top_vertices = {
  {0         , 0         , 0, 0, 255, 255, 255, 255},
  {halfWidth , halfHeight, 1, 0, 255, 255, 255, 255},
  {0         , tileHeight, 1, 1, 255, 255, 255, 255},
  {-halfWidth, halfHeight, 0, 1, 255, 255, 255, 255},
}

local left_vertices = {
  {-halfWidth, halfHeight             , 0, 0, 200, 200, 200, 255},
  {0         , tileHeight             , 1, 0, 200, 200, 200, 255},
  {0         , tileHeight * 2         , 1, 1, 200, 200, 200, 255},
  {-halfWidth, tileHeight + halfHeight, 0, 1, 200, 200, 200, 255},
}

local right_vertices = {
  {0         , tileHeight             , 0, 0, 145, 145, 145, 255},
  {halfWidth , halfHeight             , 1, 0, 145, 145, 145, 255},
  {halfWidth , tileHeight + halfHeight, 1, 1, 145, 145, 145, 255},
  {0         , tileWidth              , 0, 1, 145, 145, 145, 255},
}

local cam = {
  redraw = true,
  renderMode = "map",
  texture = texture,
  facing = 0,
  heightMod = math.random(1, 10) / 1000,
  edges = math.random(90, 110) / 100,
  dropoff = 1,
  }

function cam.setup()
  cam.resize(engine.screenWidth, engine.screenHeight)
  for i = 1, #cam.texture do
    local image = love.graphics.newImage("Data/assets/"..cam.texture[i].name..".png")
    cam.texture[i].top_mesh = love.graphics.newMesh(top_vertices, "fan", "static")
    cam.texture[i].top_mesh:setTexture(image)
    cam.texture[i].left_mesh = love.graphics.newMesh(left_vertices, "fan", "static")
    cam.texture[i].left_mesh:setTexture(image)
    cam.texture[i].right_mesh = love.graphics.newMesh(right_vertices, "fan", "static")
    cam.texture[i].right_mesh:setTexture(image)
  end
end

function cam.newMap(tiles, chunkWidth, seed, octaves)
  cam.tiles = tiles
  cam.chunkWidth = chunkWidth
  cam.seed = seed
  cam.octaves = octaves
  cam.map = {}
  --table.insert(cam.map, engine.system.chunk.new(0, 0, chunkWidth, seed, octaves, cam.texture))
  for y = -tiles, tiles do
    cam.map[y] = {}
    for x = -tiles, tiles do
      cam.map[y][x] = engine.system.chunk.new(x, y, chunkWidth, seed, octaves, cam.texture, cam.heightMod, cam.edges, cam.dropoff)
    end
  end
  cam.redraw = true
end

function cam.setFacing(f)
  cam.facing = f
end

function cam.resize(w, h)
  cam.screenWidth = w - cameraBorder
  cam.screenHeight = h
  displayCanvas = love.graphics.newCanvas(w - cameraBorder, h)
  drawCanvas = love.graphics.newCanvas(w - cameraBorder, h)
  cam.redraw = true
end

function cam.keyPressed(k)
  if k == "return" then
    cam.seed = love.math.random(1, 1000000)
    cam.newMap(cam.tiles, cam.chunkWidth, cam.seed, cam.octaves)
  elseif k == "down" then
    cam.octaves = math.min(cam.octaves + 1, 6)
    cam.newMap(cam.tiles, cam.chunkWidth, cam.seed, cam.octaves)
  elseif k == "up" then
    cam.octaves = math.max(cam.octaves - 1, 1)
    cam.newMap(cam.tiles, cam.chunkWidth, cam.seed, cam.octaves)
  elseif k == "left" then
    cam.setFacing((cam.facing - 1) % 4)
  elseif k == "right" then
    cam.setFacing((cam.facing + 1) % 4)
  end
  cam.redraw = true
end

function cam.wheelMoved(x, y, s)
  if s > 0 then
    scale = scale + zoomSpeed
  else
    scale = scale - zoomSpeed
  end
  cam.redraw = true
end

function cam.update(dt)
  if love.keyboard.isDown("w") then
    offsetY = offsetY + scrollSpeed
    cam.redraw = true
  end
  if love.keyboard.isDown("a") then
    offsetX = offsetX + scrollSpeed
    cam.redraw = true
  end
  if love.keyboard.isDown("s") then
    offsetY = offsetY - scrollSpeed
    cam.redraw = true
  end
  if love.keyboard.isDown("d") then
    offsetX = offsetX - scrollSpeed
    cam.redraw = true
  end
  cam.hoveredChunkX, cam.hoveredChunkY = cam.getHoveredChunk()
end

function cam.getChunkPosition(heading, x, y)
  --Determine the chunk screen position based on the heading specified.
  local cx, cy = 0, 0
  local displayWidth = cam.chunkWidth * halfWidth
  local displayHeight = cam.chunkWidth * halfHeight
  
  if heading == 0 then
    cx = (x - y) * displayWidth
    cy = (x + y) * displayHeight
  elseif heading == 1 then
    cx = ((cam.tiles - y - 1) * displayWidth) - (x * displayWidth)
    cy = (x - y) * displayHeight
  elseif heading == 2 then
    cx = (y * displayWidth) - (x * displayWidth)
    cy = ((cam.tiles - y) * displayHeight) - (x * displayHeight)
  elseif heading == 3 then
    cx = (y * displayWidth) + (x * displayWidth)
    cy = (cam.tiles - x + y) * displayHeight
  end
  
  return cx, cy
end

function cam.getBlockPosition(heading, x, y)
  local px, py = 0, 0
  
  if heading == 0 then
    px = (x - y) * halfWidth
    py = (x + y) * halfHeight
  elseif heading == 1 then
    px = ((halfWidth - y) - x) * halfWidth
    py = (x - y) * halfHeight
  elseif heading == 2 then
    px = (y - x) * halfWidth
    py = ((halfHeight - x) * halfHeight) - (y * halfHeight)
  elseif heading == 3 then
    px = ((y * halfWidth) + (x * halfWidth)) - (cam.chunkWidth * halfWidth)
    py = ((halfHeight - x) * halfHeight) + (y * halfHeight) + ((cam.chunkWidth * .5) * halfHeight)
  end
  
  return px, py
end

function cam.getHoveredChunk()
  local cx, cy = 0, 0
  local mx = (offsetX + engine.mousex)
  local my = (offsetY - engine.mousey)
  local invScale = scale
  local halfWidth = (cam.chunkWidth * halfWidth) / 2
  local halfHeight = (cam.chunkWidth * halfHeight) / 2
  
  if cam.facing == 0 then
    cx = math.floor(((mx / halfWidth) + (my / halfHeight)))
    cy = math.floor(((my / halfHeight) - (mx / halfWidth)) * invScale)
  end
  
  return cx, cy
end

function cam.getHoveredTile()
  local cx, cy = cam.getHoveredChunk()
  local tx, ty = 0, 0
  local mx, my = engine.mousex, engine.mousey
  local chunk = cam.map[cy][cx]
  
  cam.hoveredTileX, cam.hoveredTileY = tx, ty
end

function cam.getChunkOrder()
  local yMin, yMax, yInc, xMin, xMax, xInc = 1, 1, 1, 1, 1, 1
  if cam.facing == 0 then --NORTH
    --Chunk order
    yMin, yMax, yInc = -cam.tiles, cam.tiles, 1
    xMin, xMax, xInc = -cam.tiles, cam.tiles, 1
  elseif cam.facing == 1 then --EAST
    --Chunk order
    yMin, yMax, yInc = cam.tiles, -cam.tiles, -1
    xMin, xMax, xInc = -cam.tiles, cam.tiles, 1
  elseif cam.facing == 2 then --SOUTH
    --Chunk order
    yMin, yMax, yInc = cam.tiles, -cam.tiles, -1
    xMin, xMax, xInc = cam.tiles, -cam.tiles, -1
  elseif cam.facing == 3 then --WEST
    --Chunk order
    yMin, yMax, yInc = -cam.tiles, cam.tiles, 1
    xMin, xMax, xInc = cam.tiles, -cam.tiles, -1
  end
  return yMin, yMax, yInc, xMin, xMax, xInc
end

function cam.getTileOrder()
  local byMin, byMax, byInc = 1, 1, 1
  local bxMin, bxMax, bxInc = 1, 1, 1
  
  --Establish what order we need to go through the chunks to render them in the right order on screen.
  if cam.facing == 0 then --NORTH
    byMin, byMax, byInc = 1, cam.chunkWidth, 1
    bxMin, bxMax, bxInc = 1, cam.chunkWidth, 1
  elseif cam.facing == 1 then --EAST
    byMin, byMax, byInc = cam.chunkWidth, 1, -1
    bxMin, bxMax, bxInc = 1, cam.chunkWidth, 1
  elseif cam.facing == 2 then --SOUTH
    byMin, byMax, byInc = cam.chunkWidth, 1, -1
    bxMin, bxMax, bxInc = cam.chunkWidth, 1, -1
  elseif cam.facing == 3 then --WEST
    byMin, byMax, byInc = 1, cam.chunkWidth, 1
    bxMin, bxMax, bxInc = cam.chunkWidth, 1, -1
  end
  return byMin, byMax, byInc, bxMin, bxMax, bxInc
end

function cam.renderBlock(scrx, scry, tile, tree, invScale)
  if scrx >= -tileWidth and scrx <= (cam.screenWidth + halfWidth) *  invScale then
    if scry >= -tileHeight and scry <= cam.screenHeight * invScale then
      love.graphics.draw(cam.texture[tile].left_mesh, scrx, scry)
      love.graphics.draw(cam.texture[tile].right_mesh, scrx, scry)
      love.graphics.draw(cam.texture[tile].top_mesh, scrx, scry)
      if cam.texture[tile].name == "grass" and tree < treeChance then
        love.graphics.draw(engine.assets.trees, scrx, scry, 0, 1, 1, 16, 12)
      end
    end
  end
end

function cam.renderChunk(chunk, yMin, yMax, yInc, xMin, xMax, xInc)
  local invScale = 1 / scale
  local cx, cy = cam.getChunkPosition(cam.facing, chunk.x, chunk.y)
  for y = yMin, yMax, yInc do
    for x = xMin, xMax, xInc do
      --Get the Tile data
      local height = chunk.terrain[y][x].height * 1.5
      local tile = chunk.terrain[y][x].tile
      local tree = chunk.terrain[y][x].tree
      
      --Establish screen coords and draw
      local px, py = cam.getBlockPosition(cam.facing, x, y)
      
      --Determine facing and modify accordingly
      
      local scrx = cx + px + offsetX
      local scry = (cy + py + offsetY) - (height * 4)
      cam.renderBlock(scrx, scry, tile, tree, invScale)
    end
  end
end

function cam.renderMap(yMin, yMax, yInc, xMin, xMax, xInc)
  --Go through the chunks in the proper order and draw them to the screen
  for y = yMin, yMax, yInc do
    for x = xMin, xMax, xInc do
      local chunk = cam.map[y][x]
      local invScale = 1 / scale
      cam.renderChunk(chunk, cam.getTileOrder())
    end
  end
  
end

function cam.draw()
  if cam.redraw then
    love.graphics.setCanvas(drawCanvas)
    love.graphics.clear(0,0,0,0)
    love.graphics.push()
    love.graphics.scale(scale)
    love.graphics.setColor(255,255,255,255)
    screenWidth, screenHeight = love.graphics.getDimensions()
    if cam.renderMode == "map" then
      cam.renderMap(cam.getChunkOrder())
    elseif cam.renderMode == "chunk" then
      cam.renderChunk(cam.map[0][0], cam.getTileOrder())
    end
    love.graphics.pop()
    displayCanvas, drawCanvas = drawCanvas, displayCanvas
    cam.redraw = false
    love.graphics.setCanvas()
  end
  local fStr = "Facing: "..cam.facing
  love.graphics.print(fStr, cam.screenWidth - font:getWidth(fStr), 1)
  local posStr = "Offset: "..offsetX..", "..offsetY
  love.graphics.print(posStr, cam.screenWidth - font:getWidth(posStr), 17)
  local chunkStr = "Chunk: "..cam.hoveredChunkX..", "..cam.hoveredChunkY
  love.graphics.print(chunkStr, cam.screenWidth - font:getWidth(chunkStr), 35)
  love.graphics.draw(displayCanvas, 0, 0)
end

return cam